﻿/*
 *
 *
 *    MainForm.cs
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2017 by Matthew K. Roberts, KK5JY.
 *    All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *    
 *
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Text.RegularExpressions;
using System.Diagnostics;

using KK5JY.Log;

namespace JayLog {
	public partial class MainForm : Form {
		#region Constants
		private const int COL_DATEON = 0;
		private const int COL_DATEOFF = 1;
		private const int COL_TIMEON = 2;
		private const int COL_TIMEOFF = 3;
		private const int COL_CALL = 4;
		private const int COL_FREQ = 5;
		private const int COL_MODE = 6;
		private const int COL_RXTEXT = 7;
		private const int COL_TXTEXT = 8;
		private const int COL_NOTES = 9;

		private readonly Color GoodColor = SystemColors.Window;
		private readonly Color WarnColor = Color.Yellow;
		#endregion

		#region Inner Types
		/// <summary>
		/// Represents a historical value of a cell.
		/// </summary>
		private class CellValue {
			public DataGridViewRow Row { get; set; }
			public int Columnn { get; set; }
			public string Value { get; set; }
		}
		#endregion

		#region Private Fields
		private Timer m_Timer;
		private Regex[] m_ValidationExpressions;
		private LinkedList<CellValue> m_UndoBuffer;
		private DateTime m_LastModification;
		private string m_MostRecentPath;
		private bool m_ModCheckExclusionFlag;
		private bool m_Modified;
		private bool m_UndoMute;

		/// <summary>
		/// Get/set the grid font size.
		/// </summary>
		private float GridFontSize {
			get { return dgMain.Font.SizeInPoints; }
			set { dgMain.Font = new Font(dgMain.Font.FontFamily, value); }
		}
		#endregion

		#region Constructors
		public MainForm() {
			InitializeComponent();
			m_Timer = new Timer();
			m_Timer.Interval = 2000;
			m_Timer.Tick += TimerTick;
			m_ValidationExpressions = new Regex[10];
			m_ValidationExpressions[COL_DATEON] = new Regex("^[0-9]{8}$");
			m_ValidationExpressions[COL_DATEOFF] = new Regex("^[0-9]{8}$");
			m_ValidationExpressions[COL_TIMEON] = new Regex("^[0-9]{4,6}$");
			m_ValidationExpressions[COL_TIMEOFF] = new Regex("^[0-9]{4,6}$");
			m_ValidationExpressions[COL_CALL] = null;
			m_ValidationExpressions[COL_FREQ] = new Regex("^[0-9,.]*$");
			m_ValidationExpressions[COL_MODE] = null;
			m_ValidationExpressions[COL_RXTEXT] = null;
			m_ValidationExpressions[COL_TXTEXT] = null;
			m_ValidationExpressions[COL_NOTES] = null;
			m_UndoBuffer = new LinkedList<CellValue>();
			m_Modified = false;
			m_UndoMute = false;
			m_ModCheckExclusionFlag = false;
			GridFontSize = 12.0f;

			// setup mouse wheel zooming
			dgMain.MouseWheel += new MouseEventHandler(GridMouseWheel);
			dgMain.AutoSizeRowsMode = DataGridViewAutoSizeRowsMode.AllCells;

			// track focus events
			this.GotFocus += new EventHandler(MainFormFocused);
			this.Activated += new EventHandler(MainFormFocused);

			// update status line
			UpdateStatus();
		}
		#endregion

		#region UI Event Handlers
		/// <summary>
		/// Timer handler.
		/// </summary>
		private void TimerTick(object sender, EventArgs e) {
			CheckOutsideModification();
		}

		/// <summary>
		/// Handle focus/activate events, to check for modified file.
		/// </summary>
		private void MainFormFocused(object sender, EventArgs e) {
			Debug.WriteLine("DEBUG: MainForm got focus");
			CheckOutsideModification();
		}

		/// <summary>
		/// File | Open
		/// </summary>
		private void OpenClick(object sender, EventArgs e) {
			if (!CheckInternalModification()) {
				return;
			}

			using (var dialog = new OpenFileDialog {
				Multiselect = false,
				CheckPathExists = true,
				Filter = "ADIF Files|*.adi;*.adif|All Files|*.*",
				Title = "Open ADIF"
			}) {
				if (dialog.ShowDialog() == DialogResult.OK) {
					try {
						ReadRowsFromFile(dialog.FileName, false);
						m_MostRecentPath = dialog.FileName;
						UpdateStatus();
					} catch (Exception ex) {
						//
						//  TODO: This should be more descriptive
						//
						MessageBox.Show(ex.GetType().Name, ex.Message);
					}
				}
			}
		}

		/// <summary>
		/// File | Save As...
		/// </summary>
		private void SaveAsClick(object sender, EventArgs e) {
			using (var dialog = new SaveFileDialog {
				Filter = "ADIF Files|*.adi;*.adif|All Files|*.*",
				Title = "Save ADIF"
			}) {
				if (dialog.ShowDialog() == DialogResult.OK) {
					try {
						WriteRowsToFile(dialog.FileName);
						m_Modified = false;
					} catch (Exception ex) {
						//
						//  TODO: This should be more descriptive
						//
						MessageBox.Show(ex.GetType().Name, ex.Message);
					}
				}
			}
		}

		/// <summary>
		/// File | Save
		/// </summary>
		private void SaveClick(object sender, EventArgs e) {
			try {
				if (String.IsNullOrEmpty(m_MostRecentPath)) {
					SaveAsClick(sender, e);
					return;
				}
				WriteRowsToFile(m_MostRecentPath);
				m_Modified = false;
				UpdateStatus();
			} catch (Exception ex) {
				//
				//  TODO: This should be more descriptive
				//
				MessageBox.Show(ex.GetType().Name, ex.Message);
			}
		}

		/// <summary>
		/// File | Import...
		/// </summary>
		private void ImportClick(object sender, EventArgs e) {
			using (var dialog = new OpenFileDialog {
				Multiselect = false,
				CheckPathExists = true,
				Filter = "ADIF Files|*.adi;*.adif|All Files|*.*",
				Title = "Import ADIF"
			}) {
				if (dialog.ShowDialog() == DialogResult.OK) {
					try {
						ReadRowsFromFile(dialog.FileName, true);
					} catch (Exception ex) {
						//
						//  TODO: This should be more descriptive
						//
						MessageBox.Show(ex.GetType().Name, ex.Message);
					}
				}
			}
		}

		/// <summary>
		/// File | New
		/// </summary>
		private void NewClick(object sender, EventArgs e) {
			if (!CheckInternalModification()) {
				return;
			}

			dgMain.Rows.Clear();
			m_UndoBuffer.Clear();
			m_Modified = false;
			m_LastModification = DateTime.MinValue;
			m_MostRecentPath = null;
			UpdateStatus();
		}

		/// <summary>
		/// File | Exit
		/// </summary>
		private void ExitClick(object sender, EventArgs e) {
			this.Close();
		}

		/// <summary>
		/// Raised when form is loaded.
		/// </summary>
		protected override void OnLoad(EventArgs e) {
			// if a filename was specified, try to load it
			var args = Environment.GetCommandLineArgs();
			for (int i = 0; i != args.Length; ++i) {
				Console.WriteLine(String.Format("DEBUG: argv[{0}] = {1}", i, args[i]));
			}
			if (args.Length > 1) {
				var filename = args[1];
				if (!File.Exists(filename)) {
					using (var file = File.OpenWrite(filename)) {
						var header = new AdifHeader();
						header.WriteToFile(file);
					}
				}
				try {
					ReadRowsFromFile(filename, false);
					m_MostRecentPath = filename;
					UpdateStatus();
				} catch (Exception ex) {
					MessageBox.Show(String.Format("Unable to read file {0}: {1}", filename, ex.Message));
				}
			}

			// start the timer
			m_Timer.Start();
			base.OnLoad(e);
		}

		/// <summary>
		/// Raised when the form is closing.
		/// </summary>
		/// <param name="e"></param>
		protected override void OnClosing(CancelEventArgs e) {
			if (!CheckInternalModification()) {
				e.Cancel = true;
				return;
			}

			m_Timer.Stop();
			base.OnClosing(e);
		}

		/// <summary>
		/// Raised when a cell value changes.
		/// </summary>
		private void CellValueChanged(object sender, DataGridViewCellEventArgs e) {
			// validate the cell
			ValidateCell(e.RowIndex, e.ColumnIndex);

			// check most recent value to see if we can detect a real value change;
			//   then use that to generate an Undo buffer item
			string mostRecentValue = GetTag(e.RowIndex, e.ColumnIndex);
			if (ReferenceEquals(mostRecentValue, null)) return;
			var currentValue = GetCellValue(e.RowIndex, e.ColumnIndex);
			if (!String.Equals(currentValue, mostRecentValue) && !m_UndoMute) {
				m_Modified = true;
				var undo = new CellValue { Row = dgMain.Rows[e.RowIndex], Columnn = e.ColumnIndex, Value = mostRecentValue };
				m_UndoBuffer.AddLast(undo);
				Debug.WriteLine("DEBUG: Undo.Add({0}, {1}, {2})", undo.Row, undo.Columnn, undo.Value);
			}
		}

		/// <summary>
		/// Raised when a cell is selected.
		/// </summary>
		private void CellEnter(object sender, DataGridViewCellEventArgs e) {
			var value = GetCellValue(e.RowIndex, e.ColumnIndex);
			if (!ReferenceEquals(value, null)) {
				SetTag(e.RowIndex, e.ColumnIndex, value);
				Debug.WriteLine("DEBUG: cell[{0}, {1}] enter; value = {2}", e.RowIndex, e.ColumnIndex, value);
			}
		}

		/// <summary>
		/// Raised when a key is pressed on the grid.  Used to implement Undo and clipboard operations.
		/// </summary>
		private void GridKeyPress(object sender, KeyPressEventArgs e) {
			if (e.KeyChar == (char)(26)) {
				//
				//  Ctrl-Z handler
				//
				Debug.WriteLine("DEBUG: Ctrl-Z");
				e.Handled = true;

				// try to undo
				if (m_UndoBuffer.Count != 0) {
					var toRestore = m_UndoBuffer.Last.Value;
					m_UndoBuffer.RemoveLast();
					var cell = toRestore.Row.Cells[toRestore.Columnn];
					m_UndoMute = true;
					cell.Value = toRestore.Value;
					m_UndoMute = false;
					dgMain.ClearSelection();
					cell.Selected = true;
					Debug.WriteLine("DEBUG: Undo[{0}, {1}] => {2}", toRestore.Row, toRestore.Columnn, toRestore.Value);
				}
			} else if (e.KeyChar == (char)(3)) {
				//
				//  Ctrl-C handler
				//
				Debug.WriteLine("DEBUG: Ctrl-C");
				e.Handled = true;

				if (dgMain.SelectedCells.Count == 1) {
					var text = dgMain.SelectedCells[0].Value as string;
					if (!ReferenceEquals(text, null))
						Clipboard.SetText(text);
				}
			} else if (e.KeyChar == (char)(22)) {
				//
				//  Ctrl-V handler
				//
				Debug.WriteLine("DEBUG: Ctrl-V");
				e.Handled = true;

				Paste();
			}
		}

		/// <summary>
		/// Raised when the set of selected grid cells changes.
		/// </summary>
		private void GridSelectionChanged(object sender, EventArgs e) {
			// make sure the file hasn't been modified
			CheckOutsideModification();

			// refresh the tag on each cell with the current value, in case we need it later
			int cells = dgMain.SelectedCells.Count;
			if (cells != 0) {
				for (int i = 0; i != cells; ++i) {
					var cell = dgMain.SelectedCells[i];
					SetTag(cell.RowIndex, cell.ColumnIndex, cell.Value as string);
				}
			}
		}

		/// <summary>
		/// Edit | Copy
		/// </summary>
		private void CopyClick(object sender, EventArgs e) {
			if (dgMain.SelectedCells.Count != 1) return;
			var cell = dgMain.SelectedCells[0];
			Clipboard.SetText(GetCellValue(cell.RowIndex, cell.ColumnIndex));
		}

		/// <summary>
		/// Edit | Paste
		/// </summary>
		private void PasteClick(object sender, EventArgs e) {
			Paste();
		}

		/// <summary>
		/// Edit | Delete Row
		/// </summary>
		private void DeleteRowClick(object sender, EventArgs e) {
			if (dgMain.SelectedRows.Count == 0) return;

			// confirm with user
			if (MessageBox.Show(
				String.Format(
					"Selected row{0} will be deleted.  This action cannot be undone.\r\n\r\nContinue?",
					dgMain.SelectedRows.Count == 1 ? "" : "s"
				),
				"Confirm",
				MessageBoxButtons.YesNo,
				MessageBoxIcon.Question) == System.Windows.Forms.DialogResult.No)
				return;

			// get a list of rows to delete
			var toDelete = new LinkedList<DataGridViewRow>();
			foreach (var row in dgMain.SelectedRows) {
				toDelete.AddLast(row as DataGridViewRow);
				Debug.Assert(toDelete.Last.Value is DataGridViewRow);
			}

			// remove all undo items that mention any of the selected rows
			var toUndo = new LinkedList<CellValue>();
			foreach (var row in toDelete) {
				foreach (var undo in m_UndoBuffer.Where(u => u.Row == row)) {
					toUndo.AddLast(undo);
				}
			}
			foreach (var undo in toUndo) {
				m_UndoBuffer.Remove(undo);
			}

			// delete the selected rows
			foreach (var row in toDelete) {
				dgMain.Rows.Remove(row);
			}

			// grid modified
			m_Modified = true;
		}

		/// <summary>
		/// Raised by mouse wheel.
		/// </summary>
		private void GridMouseWheel(object sender, MouseEventArgs e) {
			// handle mouse zoom requests
			if (Control.ModifierKeys == Keys.Control) {
				if (e.Delta > 0) {
					if (GridFontSize >= 6)
						GridFontSize += 2;
				} else {
					if (GridFontSize < 96)
						GridFontSize -= 2;
				}
				// auto-size the columns
				dgMain.AutoResizeColumns(DataGridViewAutoSizeColumnsMode.AllCells);
			}
		}

		/// <summary>
		/// Raised when rows are removed.
		/// </summary>
		private void GridRowsRemoved(object sender, DataGridViewRowsRemovedEventArgs e) {
			UpdateStatus();
		}

		/// <summary>
		/// Raised when rows are added.
		/// </summary>
		private void GridRowsAdded(object sender, DataGridViewRowsAddedEventArgs e) {
			UpdateStatus();
		}

		/// <summary>
		/// Raised when ABOUT clicked.
		/// </summary>
		private void AboutClick(object sender, EventArgs e) {
			MessageBox.Show(
				@"JayLog Version 1.0 (beta)

Copyright (C) 2017 by Matt Roberts, KK5JY,
All rights reserved.

License: GNU General Public License Version 3.0.
    
This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
    
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
    
You should have received a copy of the GNU General Public License along with this program.  If not, see:

http://www.gnu.org/licenses/",
				"About",
				MessageBoxButtons.OK,
				MessageBoxIcon.Information);
		}
		#endregion

		#region Utilities
		/// <summary>
		/// Paste from the clipboard into selected cells.
		/// </summary>
		private void Paste() {
			if (Clipboard.ContainsText()) {
				int cells = dgMain.SelectedCells.Count;
				if (cells == 0) return;
				var text = Clipboard.GetText();
				m_UndoMute = true;
				for (int i = 0; i != cells; ++i) {
					var cell = dgMain.SelectedCells[i];
					cell.Value = text;
					var value = GetTag(cell.RowIndex, cell.ColumnIndex);
					if (value != null) {
						m_UndoBuffer.AddLast(
							new CellValue {
								Row = dgMain.Rows[cell.RowIndex],
								Columnn = cell.ColumnIndex,
								Value = value
							}
						);
					}
				}
				m_UndoMute = false;
			}
		}

		/// <summary>
		/// Attempt to validate a cell.
		/// </summary>
		private void ValidateCell(int row, int col) {
			if (row < 0 || col < 0) return;
			
			// fetch the cell
			var cell = dgMain[col, row];
			if (ReferenceEquals(cell, null)) return;

			// get the value of the cell
			var text = cell.Value as string ?? "";

			// check against the regex
			if (ReferenceEquals(m_ValidationExpressions[col], null)) return;
			var m = m_ValidationExpressions[col].Match(text);

			// change color to match the result
			dgMain[col, row].Style.BackColor = m.Success ? GoodColor : WarnColor;

			Debug.WriteLine("DEBUG: Validate({0}, {1}) = {2}", row, col, m.Success);
		}

		/// <summary>
		/// Read an ADIF file into the main grid.
		/// </summary>
		/// <param name="path">Full path to the file.</param>
		/// <param name="importOnly">If true, import rows; if false, replace all rows.</param>
		private void ReadRowsFromFile(string path, bool importOnly) {
			var header = new AdifHeader();
			//
			//  TODO: read the header
			//
			var adif = new AdifFile(path);
			if (ReferenceEquals(adif.ReadHeader(), null)) {
				throw new InvalidOperationException("Could not read the file header");
			}
			if (!importOnly)
				dgMain.Rows.Clear();
			m_UndoMute = true;
			foreach (var item in adif.GetAllRecords()) {
				var row = new DataGridViewRow();
				row.Cells.Add(new DataGridViewTextBoxCell { Value = item.DateOn });
				row.Cells.Add(new DataGridViewTextBoxCell { Value = item.DateOff });
				row.Cells.Add(new DataGridViewTextBoxCell { Value = item.TimeOn });
				row.Cells.Add(new DataGridViewTextBoxCell { Value = item.TimeOff });
				row.Cells.Add(new DataGridViewTextBoxCell { Value = item.Call });
				row.Cells.Add(new DataGridViewTextBoxCell { Value = item.Freq });
				row.Cells.Add(new DataGridViewTextBoxCell { Value = item.Mode });
				row.Cells.Add(new DataGridViewTextBoxCell { Value = item.RxString });
				row.Cells.Add(new DataGridViewTextBoxCell { Value = item.TxString });
				row.Cells.Add(new DataGridViewTextBoxCell { Value = item.Notes });
				dgMain.Rows.Insert(0, row);
			}
			m_UndoMute = false;
			m_Modified = importOnly;
			m_UndoBuffer.Clear();
			m_LastModification = (new FileInfo(path)).LastWriteTimeUtc;
		}

		/// <summary>
		/// Write the grid to an ADIF file.
		/// </summary>
		private void WriteRowsToFile(string path) {
			if (File.Exists(path)) {
				File.Delete(path);
			}
			if (dgMain.Rows.Count == 0) {
				var h = new AdifHeader();
				h.WriteToFile(File.Create(path));
				return;
			}
			try {
				var adif = new AdifFile(path);
				for (int row = dgMain.Rows.Count - 1; row >= 0; --row) {
					var rec = new AdifRecord();
					var rowData = dgMain.Rows[row];
					if (ReferenceEquals(rowData, null)) {
						Console.WriteLine("Error: could not fetch row {0}", row);
						continue;
					}
					var cell = rowData.Cells[COL_DATEON];
					if (!ReferenceEquals(cell, null))
						rec.DateOn = cell.Value as string;

					cell = rowData.Cells[COL_DATEOFF];
					if (!ReferenceEquals(cell, null))
						rec.DateOff = cell.Value as string;

					cell = rowData.Cells[COL_TIMEON];
					if (!ReferenceEquals(cell, null))
						rec.TimeOn = cell.Value as string;

					cell = rowData.Cells[COL_TIMEOFF];
					if (!ReferenceEquals(cell, null))
						rec.TimeOff = cell.Value as string;

					cell = rowData.Cells[COL_CALL];
					if (!ReferenceEquals(cell, null))
						rec.Call = cell.Value as string;

					cell = rowData.Cells[COL_FREQ];
					if (!ReferenceEquals(cell, null))
						rec.Freq = cell.Value as string;

					cell = rowData.Cells[COL_MODE];
					if (!ReferenceEquals(cell, null))
						rec.Mode = cell.Value as string;

					cell = rowData.Cells[COL_RXTEXT];
					if (!ReferenceEquals(cell, null))
						rec.RxString = cell.Value as string;

					cell = rowData.Cells[COL_TXTEXT];
					if (!ReferenceEquals(cell, null))
						rec.TxString = cell.Value as string;

					cell = rowData.Cells[COL_NOTES];
					if (!ReferenceEquals(cell, null))
						rec.Notes = cell.Value as string;

					adif.AddRecord(rec);
				}
				m_LastModification = (new FileInfo(path)).LastWriteTimeUtc;
			} catch (Exception ex) {
				MessageBox.Show(
					String.Format("While writing ADIF file: {0}: {1}",
						ex.Message,
						ex.GetType().Name),
					"ADIF Error");
			}
		}

		/// <summary>
		/// Get the value of a cell.
		/// </summary>
		private string GetCellValue(int row, int col) {
			if (row < 0 || col < 0) return null;
			var cell = dgMain[col, row];
			if (ReferenceEquals(cell, null)) return null;
			return cell.Value as string;
		}

		/// <summary>
		/// Set the tag on a cell.
		/// </summary>
		private void SetTag(int row, int col, string value) {
			if (row < 0 || col < 0) return;
			var cell = dgMain[col, row];
			if (ReferenceEquals(cell, null)) return;
			cell.Tag = value;
		}

		/// <summary>
		/// Get the tag on a cell.
		/// </summary>
		private string GetTag(int row, int col) {
			if (row < 0 || col < 0) return null;
			var cell = dgMain[col, row];
			if (ReferenceEquals(cell, null)) return null;
			return cell.Tag as string;
		}

		/// <summary>
		/// If the file has been modified by user, prompt the user to save the file.  This is used by
		/// exit handlers and by Open handler to determine whether to discard a modified buffer.
		/// </summary>
		/// <returns>True if the user has chosen to allow the code to continue; false if the calling code should cancel the proposed operation.</returns>
		private bool CheckInternalModification() {
			if (m_Modified) {
				switch (MessageBox.Show("File has been modified -- save?", "Save?", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question)) {
					case System.Windows.Forms.DialogResult.Yes: {
						SaveClick(this, new EventArgs());
						return true;
					}
					case System.Windows.Forms.DialogResult.Cancel: {
						return false;
					}
				}
			}
			return true;
		}

		/// <summary>
		/// Check to see if file has been modified outside the editor, and if so, prompt user to reload.
		/// </summary>
		private void CheckOutsideModification() {
			if (m_ModCheckExclusionFlag) return;
			try {
				m_ModCheckExclusionFlag = true;
				var lastMod = (new FileInfo(m_MostRecentPath)).LastWriteTimeUtc;
				if (lastMod != m_LastModification) {
					m_LastModification = lastMod;
					DialogResult result = m_Modified ? DialogResult.No : DialogResult.Yes;
					if (m_Modified) {
						result = MessageBox.Show("The file has been modified by another process, and there are local changes.\r\n\r\nReload file?", "File Modified",
							MessageBoxButtons.YesNo, MessageBoxIcon.Question);
					}
					if (result == DialogResult.Yes) {
						ReadRowsFromFile(m_MostRecentPath, false);
					}
				}
			} catch (Exception ex) {
				Debug.WriteLine("DEBUG: {0}: {1}", ex.GetType().Name, ex.Message);
			} finally {
				m_ModCheckExclusionFlag = false;
			}
		}

		/// <summary>
		/// Updates the status line.
		/// </summary>
		private void UpdateStatus() {
			gridStatus.Text = String.Format("{0} Records", dgMain.Rows.Count);
			if (!String.IsNullOrEmpty(m_MostRecentPath)) {
				this.Text = String.Format("JayLog - {0}", Path.GetFileName(m_MostRecentPath));
			} else {
				this.Text = "JayLog";
			}
		}
		#endregion
	}
}
